﻿#include "pch.h"
#include "BitmapTransformExperimentMain.h"
#include "Common\DirectXHelper.h"

using namespace BitmapTransformExperiment;
using namespace Windows::Foundation;
using namespace Windows::Foundation::Collections;
using namespace Windows::System::Threading;
using namespace Concurrency;

// Loads and initializes application assets when the application is loaded.
BitmapTransformExperimentMain::BitmapTransformExperimentMain(const std::shared_ptr<DX::DeviceResources>& deviceResources) :
    m_deviceResources(deviceResources), m_pointerLocationX(0.0f)
{
    // Register to be notified if the Device is lost or recreated
    m_deviceResources->RegisterDeviceNotify(this);

    // Content initialization.
    m_bitmapTransformExperimentRenderer = std::unique_ptr<BitmapTransformExperimentRenderer>(new BitmapTransformExperimentRenderer(m_deviceResources));

    // TODO: Change the timer settings if you want something other than the default variable timestep mode.
    // e.g. for 60 FPS fixed timestep update logic, call:
    /*
    m_timer.SetFixedTimeStep(true);
    m_timer.SetTargetElapsedSeconds(1.0 / 60);
    */
}

BitmapTransformExperimentMain::~BitmapTransformExperimentMain()
{
    // Deregister device notification
    m_deviceResources->RegisterDeviceNotify(nullptr);
}

// Updates application state when the window size changes (e.g. device orientation change)
void BitmapTransformExperimentMain::CreateWindowSizeDependentResources() 
{
    // TODO: Replace this with the size-dependent initialization of your app's content.
    m_bitmapTransformExperimentRenderer->CreateWindowSizeDependentResources();
}

void BitmapTransformExperimentMain::StartRenderLoop()
{
    // If the animation render loop is already running then do not start another thread.
    if (m_renderLoopWorker != nullptr && m_renderLoopWorker->Status == AsyncStatus::Started)
    {
        return;
    }

    // Create a task that will be run on a background thread.
    auto workItemHandler = ref new WorkItemHandler([this](IAsyncAction ^ action)
    {
        // Calculate the updated frame and render once per vertical blanking interval.
        while (action->Status == AsyncStatus::Started)
        {
            critical_section::scoped_lock lock(m_criticalSection);
            Update();
            if (Render())
            {
                m_deviceResources->Present();
            }
        }
    });

    // Run task on a dedicated high priority background thread.
    m_renderLoopWorker = ThreadPool::RunAsync(workItemHandler, WorkItemPriority::High, WorkItemOptions::TimeSliced);
}

void BitmapTransformExperimentMain::StopRenderLoop()
{
    m_renderLoopWorker->Cancel();
}

// Updates the application state once per frame.
void BitmapTransformExperimentMain::Update() 
{
    ProcessInput();

    // Update scene objects.
    m_timer.Tick([&]()
    {
        // Content update functions.
        m_bitmapTransformExperimentRenderer->Update(m_timer);
    });
}

// Process all input from the user before updating game state
void BitmapTransformExperimentMain::ProcessInput()
{
    // TODO: Add per frame input handling here.
}

// Renders the current frame according to the current application state.
// Returns true if the frame was rendered and is ready to be displayed.
bool BitmapTransformExperimentMain::Render() 
{
    // Don't try to render anything before the first Update.
    if (m_timer.GetFrameCount() == 0)
    {
        return false;
    }

    // Content rendering functions.
    return m_bitmapTransformExperimentRenderer->Render();
}

// Notifies renderers that device resources need to be released.
void BitmapTransformExperimentMain::OnDeviceLost()
{
    m_bitmapTransformExperimentRenderer->ReleaseDeviceDependentResources();
}

// Notifies renderers that device resources may now be recreated.
void BitmapTransformExperimentMain::OnDeviceRestored()
{
    m_bitmapTransformExperimentRenderer->CreateDeviceDependentResources();
    CreateWindowSizeDependentResources();
}

void BitmapTransformExperimentMain::MatrixValueChanged(IVector<float>^ matrix)
{
    m_bitmapTransformExperimentRenderer->MatrixValueChanged(matrix);
}