﻿#include "pch.h"
#include "BitmapTransformExperimentRenderer.h"

#include "Common/DirectXHelper.h"

using namespace BitmapTransformExperiment;

using namespace D2D1;
using namespace Microsoft::WRL;

BitmapTransformExperimentRenderer::BitmapTransformExperimentRenderer(const std::shared_ptr<DX::DeviceResources>& deviceResources) : 
    m_deviceResources(deviceResources)
{
    // Initialize matrix
    for (int row = 0; row < 4; row++)
    for (int col = 0; col < 4; col++)
        m_matrix.m[row][col] = row == col ? 1.0f : 0.0f;

    DX::ThrowIfFailed(
        m_deviceResources->GetD2DFactory()->CreateDrawingStateBlock(&m_stateBlock)
        );

    CreateDeviceDependentResources();
}

void BitmapTransformExperimentRenderer::CreateDeviceDependentResources()
{
    ID2D1DeviceContext1* context = m_deviceResources->GetD2DDeviceContext();

    DX::ThrowIfFailed(
        context->CreateSolidColorBrush(ColorF(ColorF::Red), &m_axisBrush)
        );

    IWICImagingFactory2* wicImagingFactory = m_deviceResources->GetWicImagingFactory();

    ComPtr<IWICBitmapDecoder> wicBitmapDecoder;

    DX::ThrowIfFailed(
        wicImagingFactory->CreateDecoderFromFilename(L"Bitmaps/PetzoldJersey.jpg", 
                                                     nullptr, 
                                                     GENERIC_READ, 
                                                     WICDecodeMetadataCacheOnDemand, 
                                                     &wicBitmapDecoder)
        );

    ComPtr<IWICBitmapFrameDecode> wicBitmapFrame;

    DX::ThrowIfFailed(
        wicBitmapDecoder->GetFrame(0, &wicBitmapFrame)
        );

    ComPtr<IWICFormatConverter> wicFormatConverter;

    DX::ThrowIfFailed(
        wicImagingFactory->CreateFormatConverter(&wicFormatConverter)
        );

    DX::ThrowIfFailed(
        wicFormatConverter->Initialize(wicBitmapFrame.Get(),
            GUID_WICPixelFormat32bppPBGRA,
            WICBitmapDitherTypeNone,
            nullptr,
            0,
            WICBitmapPaletteTypeCustom)
        );

    // Set bitmap DPI to 96
    float dpiX = 96, dpiY = 96;

    DX::ThrowIfFailed(
        context->CreateBitmapFromWicBitmap(wicFormatConverter.Get(),
                                           BitmapProperties(PixelFormat(DXGI_FORMAT_B8G8R8A8_UNORM, 
                                                                        D2D1_ALPHA_MODE_PREMULTIPLIED),
                                                            dpiX, 
                                                            dpiY),
                                           &m_bitmap)
    );

    m_needsRedraw = true;
}
void BitmapTransformExperimentRenderer::ReleaseDeviceDependentResources()
{
    m_axisBrush.Reset();
    m_bitmap.Reset();
}

void BitmapTransformExperimentRenderer::CreateWindowSizeDependentResources()
{
    m_needsRedraw = true;
}

void BitmapTransformExperimentRenderer::Update(DX::StepTimer const& timer)
{
}

// Renders a frame to the screen.
bool BitmapTransformExperimentRenderer::Render()
{
    if (!m_needsRedraw)
        return false;

    ID2D1DeviceContext* context = m_deviceResources->GetD2DDeviceContext();
    Windows::Foundation::Size logicalSize = m_deviceResources->GetLogicalSize();
    D2D1_SIZE_F bitmapSize = m_bitmap->GetSize();

    context->SaveDrawingState(m_stateBlock.Get());
    context->BeginDraw();
    context->Clear(ColorF(ColorF::Black));

    Matrix3x2F centerTranslation = Matrix3x2F::Translation(
                    (logicalSize.Width - bitmapSize.width) / 2, 
                    (logicalSize.Height - bitmapSize.height) / 2);

    context->SetTransform(centerTranslation * m_deviceResources->GetOrientationTransform2D());

    // Draw axis lines
    context->DrawLine(Point2F((bitmapSize.width - logicalSize.Width) / 2, 0), 
                      Point2F((bitmapSize.width + logicalSize.Width) / 2, 0), m_axisBrush.Get());

    context->DrawLine(Point2F(0, (bitmapSize.height -logicalSize.Height) / 2), 
                      Point2F(0, (bitmapSize.height + logicalSize.Height) / 2), m_axisBrush.Get());

    for (int x =  (int) ((bitmapSize.width - logicalSize.Width) / 200); 
             x <= (int) ((bitmapSize.width + logicalSize.Width) / 200); x++)
    {
        context->DrawLine(Point2F(100.0f * x, -10.0f), Point2F(100.0f * x, 10.0f), m_axisBrush.Get());
    }

    for (int y =  (int) ((bitmapSize.height - logicalSize.Height) / 200); 
             y <= (int) ((bitmapSize.height + logicalSize.Height) / 200); y++)
    {
        context->DrawLine(Point2F(-10.0f, 100.0f * y), Point2F(10.0f, 100.0f * y), m_axisBrush.Get());
    }

    // Draw bitmap
    context->DrawBitmap(m_bitmap.Get(), 
                        nullptr, 
                        1.0f, 
                        D2D1_INTERPOLATION_MODE_LINEAR, 
                        nullptr, 
                        &m_matrix);

    // Ignore D2DERR_RECREATE_TARGET here. This error indicates that the device
    // is lost. It will be handled during the next call to Present.
    HRESULT hr = context->EndDraw();
    if (hr != D2DERR_RECREATE_TARGET)
    {
        DX::ThrowIfFailed(hr);
    }

    context->RestoreDrawingState(m_stateBlock.Get());

    m_needsRedraw = false;
    return true;
}

void BitmapTransformExperimentRenderer::MatrixValueChanged(Windows::Foundation::Collections::IVector<float>^ matrix)
{
    for (unsigned int i = 0; i < matrix->Size; i++)
        m_matrix.m[i / 4][i % 4] = matrix->GetAt(i);

    m_needsRedraw = true;
}