﻿#include "pch.h"
#include "NonAffineStretchRenderer.h"

#include "Common/DirectXHelper.h"

using namespace NonAffineStretch;

using namespace D2D1;
using namespace Microsoft::WRL;
using namespace Platform;

NonAffineStretchRenderer::NonAffineStretchRenderer(const std::shared_ptr<DX::DeviceResources>& deviceResources,
                                                   IMatrixDisplay^ matrixDisplay) : 
    m_deviceResources(deviceResources),
    m_matrixDisplay(matrixDisplay),
    m_needsRedraw(true)
{
    DX::ThrowIfFailed(
        m_deviceResources->GetD2DFactory()->CreateDrawingStateBlock(&m_stateBlock)
        );

    CreateDeviceDependentResources();
}

void NonAffineStretchRenderer::CreateDeviceDependentResources()
{
    DX::ThrowIfFailed(
        m_deviceResources->GetD2DDeviceContext()->CreateSolidColorBrush(ColorF(ColorF::Green), &m_thumbBrush)
        );

    IWICImagingFactory2* wicImagingFactory = m_deviceResources->GetWicImagingFactory();

    ComPtr<IWICBitmapDecoder> wicBitmapDecoder;

    DX::ThrowIfFailed(
        wicImagingFactory->CreateDecoderFromFilename(L"Bitmaps/PetzoldJersey.jpg", 
                                                     nullptr, 
                                                     GENERIC_READ, 
                                                     WICDecodeMetadataCacheOnDemand, 
                                                     &wicBitmapDecoder)
        );

    ComPtr<IWICBitmapFrameDecode> wicBitmapFrame;

    DX::ThrowIfFailed(
        wicBitmapDecoder->GetFrame(0, &wicBitmapFrame)
        );

    ComPtr<IWICFormatConverter> wicFormatConverter;

    DX::ThrowIfFailed(
        wicImagingFactory->CreateFormatConverter(&wicFormatConverter)
        );

    DX::ThrowIfFailed(
        wicFormatConverter->Initialize(wicBitmapFrame.Get(),
        GUID_WICPixelFormat32bppPBGRA,
        WICBitmapDitherTypeNone,
        nullptr,
        0,
        WICBitmapPaletteTypeCustom)
        );


    float dpiX = 96, dpiY = 96;

    DX::ThrowIfFailed(
        m_deviceResources->GetD2DDeviceContext()->CreateBitmapFromWicBitmap(wicFormatConverter.Get(),
                                                                            BitmapProperties1(D2D1_BITMAP_OPTIONS_NONE, 
                                                                                              PixelFormat(DXGI_FORMAT_B8G8R8A8_UNORM, 
                                                                                                          D2D1_ALPHA_MODE_PREMULTIPLIED),
                                                                                              dpiX, dpiY),
                                                                            &m_bitmap)
        );

    D2D1_SIZE_F size = m_bitmap->GetSize();

    // Initialize the Thumbs array
    m_thumbs[0].center = Point2F(100, 100);
    m_thumbs[1].center = Point2F(100 + size.width, 100);
    m_thumbs[2].center = Point2F(100, 100 + size.height);
    m_thumbs[3].center = Point2F(100 + size.width, 100 + size.height);

    for (Thumb& thumb : m_thumbs)
    {
        thumb.isTracking = false;
    }

    CalculateNewTransform();
    m_needsRedraw = true;
}
void NonAffineStretchRenderer::ReleaseDeviceDependentResources()
{
    m_thumbBrush.Reset();
}

void NonAffineStretchRenderer::CreateWindowSizeDependentResources()
{
    m_displayTransform = m_deviceResources->GetOrientationTransform2D();

    // And get the inverse matrix for hit-testing
    m_inverseDisplayTransform = m_displayTransform;
    m_inverseDisplayTransform.Invert();

    m_needsRedraw = true;
}

void NonAffineStretchRenderer::Update(DX::StepTimer const& timer)
{
}

bool NonAffineStretchRenderer::Render()
{
    if (!m_needsRedraw)
        return false;

    ID2D1DeviceContext* context = m_deviceResources->GetD2DDeviceContext();

    context->SaveDrawingState(m_stateBlock.Get());
    context->BeginDraw();
    context->Clear(ColorF(ColorF::DarkBlue));
    context->SetTransform(m_displayTransform);

    // Draw the bitmap
    context->DrawBitmap(m_bitmap.Get(), 
                        nullptr, 
                        1.0f, 
                        D2D1_INTERPOLATION_MODE_LINEAR, 
                        nullptr, 
                        &m_matrix);

    // Draw the thumbs
    for (Thumb thumb : m_thumbs)
    {
        D2D1_ELLIPSE ellipse = Ellipse(thumb.center, 
                                       (float) m_thumbRadius, 
                                       (float) m_thumbRadius);

        context->FillEllipse(ellipse, m_thumbBrush.Get());
    }

    // Ignore D2DERR_RECREATE_TARGET here. This error indicates that the device
    // is lost. It will be handled during the next call to Present.
    HRESULT hr = context->EndDraw();
    if (hr != D2DERR_RECREATE_TARGET)
    {
        DX::ThrowIfFailed(hr);
    }

    context->RestoreDrawingState(m_stateBlock.Get());

    m_needsRedraw = false;
    return true;
}

void NonAffineStretchRenderer::OnPointerPressed(unsigned int id, D2D1_POINT_2F pointerPos)
{
    // Get the iterator for pointerTracker
    auto const it = pointerTracker.find(id);

    // If this iterator already exists, remove it from the map
    if (it != pointerTracker.end())
    {
        OutputDebugString(L"OnPointerPressed without corresponding OnPointerReleased");
        pointerTracker.erase(it);
    }

    pointerPos = m_inverseDisplayTransform.TransformPoint(pointerPos);
    bool haveHit = false;
    PointerInfo pointerInfo;

    for (int offset = 1; offset <= m_thumbRadius; offset++)
    {
        for (size_t thumbIndex = 0; thumbIndex < m_thumbs.size(); thumbIndex++)
        {
            Thumb& thumb = m_thumbs[thumbIndex];
            haveHit = HitTest(pointerPos, thumb.center, (float)offset);

            if (haveHit)
            {
                pointerInfo.thumbIndex = thumbIndex;
                thumb.isTracking = true;
                break;
            }
        }

        if (haveHit)
        {
            // Set the remaining fields of the PointerInfo structure
            pointerInfo.lastPointerPosition = pointerPos;

            // Store in map with ID number
            pointerTracker.insert(std::make_pair(id, pointerInfo));
            break;
        }

    }
}

void NonAffineStretchRenderer::OnPointerMoved(unsigned int id, D2D1_POINT_2F pointerPos)
{
    // Get the iterator for pointerTracker
    auto const it = pointerTracker.find(id);

    // Never started a pointer operation, so skip
    if (it == pointerTracker.end())
        return;

    // Get the PointerInfo object for this finger
    pointerPos = m_inverseDisplayTransform.TransformPoint(pointerPos);
    PointerInfo& pointerInfo = it->second;
    MoveThumb(pointerInfo, pointerPos);

    CalculateNewTransform();
}

void NonAffineStretchRenderer::OnPointerReleased(unsigned int id, D2D1_POINT_2F pointerPos)
{
    // Get the iterator for pointerTracker
    auto const it = pointerTracker.find(id);

    // Never started a pointer operation, so skip
    if (it == pointerTracker.end())
        return;

    // Get the PointerInfo object for this finger
    pointerPos = m_inverseDisplayTransform.TransformPoint(pointerPos);
    PointerInfo& pointerInfo = it->second;
    MoveThumb(pointerInfo, pointerPos);

    // Remove from map
    Thumb& thumb = m_thumbs[pointerInfo.thumbIndex];
    thumb.isTracking = false;
    pointerTracker.erase(it);

    CalculateNewTransform();
}

void NonAffineStretchRenderer::MoveThumb(PointerInfo& pointerInfo, D2D1_POINT_2F pointerPos)
{
    // Translate thumb, and save current position as last position
    D2D1_POINT_2F& center = m_thumbs[pointerInfo.thumbIndex].center;
    center.x += pointerPos.x - pointerInfo.lastPointerPosition.x;
    center.y += pointerPos.y - pointerInfo.lastPointerPosition.y;
    pointerInfo.lastPointerPosition = pointerPos;
}

void NonAffineStretchRenderer::CalculateNewTransform()
{
    m_matrix = CalculateNewTransform(m_bitmap->GetSize(),
                                     m_thumbs[0].center,
                                     m_thumbs[1].center,
                                     m_thumbs[2].center,
                                     m_thumbs[3].center);

    m_matrixDisplay->SetMatrix(ref new Array<float>((float *)m_matrix.m, 16));

    m_needsRedraw = true;
}

// See "Programming Windows," 6th edition by Charles Petzold,
//      Chapter 10 for method below

// The returned transform maps the points (0, 0),
//  (0, 1), (1, 0), and (1, 1) to the points 
//  ptUL, ptUR, ptLL, and ptLR normalized based on size.
D2D1_MATRIX_4X4_F NonAffineStretchRenderer::CalculateNewTransform(D2D1_SIZE_F size,
                                                                  D2D1_POINT_2F ptUL,
                                                                  D2D1_POINT_2F ptUR,
                                                                  D2D1_POINT_2F ptLL,
                                                                  D2D1_POINT_2F ptLR)
{
    // Scale and translate normalization transform
    Matrix4x4F S = Matrix4x4F();
    S._11 = 1 / size.width;
    S._22 = 1 / size.height;
    S._33 = 1;
    S._44 = 1;

    // Affine transform: Maps
    //      (0, 0) --> ptUL
    //      (1, 0) --> ptUR
    //      (0, 1) --> ptLL
    //      (1, 1) --> (x2 + x1 + x0, y2 + y1 + y0)
    Matrix4x4F A = Matrix4x4F();
    A._41 = ptUL.x;
    A._42 = ptUL.y;
    A._11 = ptUR.x - ptUL.x;
    A._12 = ptUR.y - ptUL.y;
    A._21 = ptLL.x - ptUL.x;
    A._22 = ptLL.y - ptUL.y;
    A._33 = 1;
    A._44 = 1;

    // Non-Affine transform
    Matrix4x4F B = Matrix4x4F();
    float den = A._11 * A._22 - A._12 * A._21;
    float a = (A._22 * ptLR.x - A._21 * ptLR.y +
        A._21 * A._42 - A._22 * A._41) / den;
    float b = (A._11 * ptLR.y - A._12 * ptLR.x +
        A._12 * A._41 - A._11 * A._42) / den;

    B._11 = a / (a + b - 1);
    B._22 = b / (a + b - 1);
    B._14 = B._11 - 1;
    B._24 = B._22 - 1;
    B._33 = 1;
    B._44 = 1;

    // Product of three transforms
    return S * B * A;
}
