﻿#include "pch.h"
#include "NonAffineStretchMain.h"
#include "Common\DirectXHelper.h"

using namespace NonAffineStretch;

using namespace Concurrency;
using namespace D2D1;                   // for Point2F
using namespace Windows::Foundation;    // for Point
using namespace Windows::System::Threading;
using namespace Windows::UI::Core;      // for PointerEventArgs


// Loads and initializes application assets when the application is loaded.
NonAffineStretchMain::NonAffineStretchMain(const std::shared_ptr<DX::DeviceResources>& deviceResources,
                                           IMatrixDisplay^ matrixDisplay) :
    m_deviceResources(deviceResources), m_pointerLocationX(0.0f)
{
    // Register to be notified if the Device is lost or recreated
    m_deviceResources->RegisterDeviceNotify(this);

    // Content initialization.
    m_nonAffineStretchRenderer = std::unique_ptr<NonAffineStretchRenderer>(new NonAffineStretchRenderer(m_deviceResources, matrixDisplay));

    // TODO: Change the timer settings if you want something other than the default variable timestep mode.
    // e.g. for 60 FPS fixed timestep update logic, call:
    /*
    m_timer.SetFixedTimeStep(true);
    m_timer.SetTargetElapsedSeconds(1.0 / 60);
    */
}

NonAffineStretchMain::~NonAffineStretchMain()
{
    // Deregister device notification
    m_deviceResources->RegisterDeviceNotify(nullptr);
}

// Updates application state when the window size changes (e.g. device orientation change)
void NonAffineStretchMain::CreateWindowSizeDependentResources() 
{
    // Size-dependent initialization of your app's content.
    m_nonAffineStretchRenderer->CreateWindowSizeDependentResources();
}

void NonAffineStretchMain::StartRenderLoop()
{
    // If the animation render loop is already running then do not start another thread.
    if (m_renderLoopWorker != nullptr && m_renderLoopWorker->Status == AsyncStatus::Started)
    {
        return;
    }

    // Create a task that will be run on a background thread.
    auto workItemHandler = ref new WorkItemHandler([this](IAsyncAction ^ action)
    {
        // Calculate the updated frame and render once per vertical blanking interval.
        while (action->Status == AsyncStatus::Started)
        {
            critical_section::scoped_lock lock(m_criticalSection);
            Update();
            if (Render())
            {
                m_deviceResources->Present();
            }
        }
    });

    // Run task on a dedicated high priority background thread.
    m_renderLoopWorker = ThreadPool::RunAsync(workItemHandler, WorkItemPriority::High, WorkItemOptions::TimeSliced);
}

void NonAffineStretchMain::StopRenderLoop()
{
    m_renderLoopWorker->Cancel();
}

// Updates the application state once per frame.
void NonAffineStretchMain::Update() 
{
    ProcessInput();

    // Update scene objects.
    m_timer.Tick([&]()
    {
        // Content update functions.
        m_nonAffineStretchRenderer->Update(m_timer);
    });
}

// Process all input from the user before updating game state
void NonAffineStretchMain::ProcessInput()
{
    // TODO: Add per frame input handling here.
}

// Renders the current frame according to the current application state.
// Returns true if the frame was rendered and is ready to be displayed.
bool NonAffineStretchMain::Render() 
{
    // Don't try to render anything before the first Update.
    if (m_timer.GetFrameCount() == 0)
    {
        return false;
    }

    // Render the scene objects.
    // Content rendering functions.
    return m_nonAffineStretchRenderer->Render();
}

// Notifies renderers that device resources need to be released.
void NonAffineStretchMain::OnDeviceLost()
{
    m_nonAffineStretchRenderer->ReleaseDeviceDependentResources();
}

// Notifies renderers that device resources may now be recreated.
void NonAffineStretchMain::OnDeviceRestored()
{
    m_nonAffineStretchRenderer->CreateDeviceDependentResources();
    CreateWindowSizeDependentResources();
}

void NonAffineStretchMain::OnPointerPressed(unsigned int id, Point point)
{
    m_nonAffineStretchRenderer->OnPointerPressed(id, Point2F(point.X, point.Y));
}

void NonAffineStretchMain::OnPointerMoved(unsigned int id, Point point)
{
    m_nonAffineStretchRenderer->OnPointerMoved(id, Point2F(point.X, point.Y));
}

void NonAffineStretchMain::OnPointerReleased(unsigned int id, Point point)
{
    m_nonAffineStretchRenderer->OnPointerReleased(id, Point2F(point.X, point.Y));
}

