﻿#include "pch.h"
#include "RotatingTextRenderer.h"

#include "Common/DirectXHelper.h"

using namespace RotatingText;

using namespace D2D1;
using namespace Microsoft::WRL;

RotatingTextRenderer::RotatingTextRenderer(const std::shared_ptr<DX::DeviceResources>& deviceResources) : 
    m_deviceResources(deviceResources)
{
    DX::ThrowIfFailed(
        m_deviceResources->GetD2DFactory()->CreateDrawingStateBlock(&m_stateBlock)
        );

    CreateDeviceDependentResources();
}

void RotatingTextRenderer::CreateDeviceDependentResources()
{

}

void RotatingTextRenderer::CreateWindowSizeDependentResources()
{
    float displayWidth = m_deviceResources->GetLogicalSize().Width;
    IDWriteFactory2 * dwriteFactory = m_deviceResources->GetDWriteFactory();

    // Create TextFormat object
    ComPtr<IDWriteTextFormat> textFormat;

    DX::ThrowIfFailed(
        dwriteFactory->CreateTextFormat(L"Times New Roman",
                                        nullptr,
                                        DWRITE_FONT_WEIGHT_NORMAL,
                                        DWRITE_FONT_STYLE_NORMAL,
                                        DWRITE_FONT_STRETCH_NORMAL,

                                        // empirical to be half the screen width
                                        0.13f * displayWidth, 
                                        L"",
                                        &textFormat)
        );

    // Create Text Layout object
    ComPtr<IDWriteTextLayout> textLayout;

    DX::ThrowIfFailed(
        dwriteFactory->CreateTextLayout(L"ROTATE", 
                                        6, 
                                        textFormat.Get(), 
                                        INFINITY, 
                                        INFINITY, 
                                        &textLayout)
        );

    // Get the size of the formatted text
    DWRITE_TEXT_METRICS textMetrics;

    DX::ThrowIfFailed(
        textLayout->GetMetrics(&textMetrics)
        );

    // Create a compatible bitmap render target the size of the text
    ID2D1DeviceContext1* context = m_deviceResources->GetD2DDeviceContext();
    ComPtr<ID2D1BitmapRenderTarget> bitmapRenderTarget;

    DX::ThrowIfFailed(
        context->CreateCompatibleRenderTarget(SizeF(textMetrics.width, 
                                                    textMetrics.height), 
                                              &bitmapRenderTarget)
        );

    // Create a solid color brush
    ComPtr<ID2D1SolidColorBrush> solidColorBrush;

    DX::ThrowIfFailed(
        context->CreateSolidColorBrush(ColorF(ColorF::LightCyan), 
                                       &solidColorBrush)
        );;

    // Render the text on the bitmap
    bitmapRenderTarget->BeginDraw();

    bitmapRenderTarget->DrawTextLayout(Point2F(0, 0), 
                                       textLayout.Get(), 
                                       solidColorBrush.Get());
    bitmapRenderTarget->EndDraw();

    // Extract the bitmap; throw everything else away
    DX::ThrowIfFailed(
        bitmapRenderTarget->GetBitmap(&m_bitmap)
        );
}

void RotatingTextRenderer::ReleaseDeviceDependentResources()
{
    m_bitmap.Reset();
}

void RotatingTextRenderer::Update(DX::StepTimer const& timer)
{
    if (m_bitmap == nullptr)
        return;

    // Begin with the identity transform
    m_matrix = Matrix4x4F();

    // Rotate around the Y axis
    double seconds = timer.GetTotalSeconds();
    float angle = 360 * float(fmod(seconds, 7) / 7);
    m_matrix = m_matrix * Matrix4x4F::RotationY(angle);

    // Apply perspective based on the bitmap width
    D2D1_SIZE_F bitmapSize = m_bitmap->GetSize();
    m_matrix = m_matrix * Matrix4x4F::PerspectiveProjection(bitmapSize.width);
}

void RotatingTextRenderer::Render()
{
    if (m_bitmap == nullptr)
        return;

    ID2D1DeviceContext* context = m_deviceResources->GetD2DDeviceContext();
    Windows::Foundation::Size logicalSize = m_deviceResources->GetLogicalSize();

    context->SaveDrawingState(m_stateBlock.Get());
    context->BeginDraw();
    context->Clear(ColorF(ColorF::DarkMagenta));

    // Move origin to top center of screen
    Matrix3x2F centerTranslation = 
                Matrix3x2F::Translation(logicalSize.Width / 2, 0);

    context->SetTransform(centerTranslation * 
                m_deviceResources->GetOrientationTransform2D());

    // Draw the bitmap
    context->DrawBitmap(m_bitmap.Get(), 
                        nullptr, 
                        1.0f, 
                        D2D1_INTERPOLATION_MODE_LINEAR, 
                        nullptr, 
                        &m_matrix);

    // Ignore D2DERR_RECREATE_TARGET here. This error indicates that the device
    // is lost. It will be handled during the next call to Present.
    HRESULT hr = context->EndDraw();
    if (hr != D2DERR_RECREATE_TARGET)
    {
        DX::ThrowIfFailed(hr);
    }

    context->RestoreDrawingState(m_stateBlock.Get());
}
